/****************************************************************************
**
** Copyright (C) 2012 Digia Plc and/or its subsidiary(-ies).
** Contact: http://www.qt-project.org/legal
**
** This file is part of Qt Creator.
**
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Digia.  For licensing terms and
** conditions see http://qt.digia.com/licensing.  For further information
** use the contact form at http://qt.digia.com/contact-us.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Digia gives you certain additional
** rights.  These rights are described in the Digia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
****************************************************************************/

#include "merconnectionmanager.h"
#include "merconstants.h"

#include <coreplugin/icore.h>
#include <ssh/sshremoteprocessrunner.h>
#include <ssh/sshkeygenerator.h>
#include <utils/fileutils.h>

#include <QDesktopServices>
#include <QFileInfo>

using namespace QSsh;

namespace Mer {
namespace Internal {

MerConnectionManager *MerConnectionManager::m_connectionManager = 0;

MerConnectionManager::MerConnectionManager()
    : m_sshConnectionManager(SshConnectionManager::instance()),
      m_sshRemoteProcessRunner(new SshRemoteProcessRunner())
{
    m_sshConnectionParameters.host = QLatin1String("localhost");
    m_sshConnectionParameters.userName = QLatin1String(Constants::MER_SDK_USERNAME);
    m_sshConnectionParameters.authenticationType = SshConnectionParameters::AuthenticationByKey;
    m_sshConnectionParameters.proxyType = SshConnectionParameters::NoProxy;
    m_sshConnectionParameters.timeout = 10;
    m_sshConnectionParameters.privateKeyFile =
            QDesktopServices::storageLocation(QDesktopServices::HomeLocation) +
            QLatin1String("/.ssh/mer-qt-creator-rsa");

    connect(m_sshRemoteProcessRunner, SIGNAL(connectionError()),
            this, SIGNAL(remoteProcessConnectionError()));
    connect(m_sshRemoteProcessRunner, SIGNAL(processStarted()),
            this, SIGNAL(remoteProcessStarted()));
    connect(m_sshRemoteProcessRunner, SIGNAL(readyReadStandardOutput()),
            this, SIGNAL(remoteProcessReadReadyStandardOutput()));
    connect(m_sshRemoteProcessRunner, SIGNAL(readyReadStandardError()),
            this, SIGNAL(remoteProcessReadReadyStandardError()));
    connect(m_sshRemoteProcessRunner, SIGNAL(processClosed(int)),
            this, SIGNAL(remoteProcessClosed(int)));
}

MerConnectionManager::~MerConnectionManager()
{
    delete m_sshRemoteProcessRunner;
}

MerConnectionManager *MerConnectionManager::instance()
{
    if (m_connectionManager)
        return m_connectionManager;

    m_connectionManager = new MerConnectionManager;
    return m_connectionManager;
}

SshConnection *MerConnectionManager::acquireConnection()
{
    SshConnection *connection = m_sshConnectionManager.acquireConnection(m_sshConnectionParameters);
    if (connection->state() == QSsh::SshConnection::Unconnected)
        connection->connectToHost();
    return connection;
}

void MerConnectionManager::releaseConnection(SshConnection *connection)
{
    if (connection)
        m_sshConnectionManager.releaseConnection(connection);
}

void MerConnectionManager::executeCommand(const QByteArray &command)
{
    m_sshRemoteProcessRunner->run(command, m_sshConnectionParameters);
}

SshConnectionParameters MerConnectionManager::sshConnectionParameters() const
{
    return m_sshConnectionParameters;
}

void MerConnectionManager::setSshConnectionParameters(const SshConnectionParameters &params)
{
    m_sshConnectionParameters = params;
}

int MerConnectionManager::createAndAuthorizeSshKeyIfNeeded(const QString &sharedHome,
                                                            QString &errorMessage)
{
    const QString privKeyPath = m_sshConnectionParameters.privateKeyFile;
    const QString pubKeyPath = privKeyPath + QLatin1String(".pub");
    const QString authorizedKeysPath = sharedHome + QLatin1String("/.ssh/authorized_keys");
    const QFileInfo fi(privKeyPath);

    bool success = false;
    QByteArray publicKey;
    if (!fi.exists()) {
        QSsh::SshKeyGenerator keyGen;
        success = keyGen.generateKeys(QSsh::SshKeyGenerator::Rsa,
                                           QSsh::SshKeyGenerator::Mixed, 2048,
                                           QSsh::SshKeyGenerator::DoNotOfferEncryption);
        if (!success) {
            errorMessage = keyGen.error();
            return KeyGenError;
        }

        Utils::FileSaver privKeySaver(privKeyPath);
        privKeySaver.write(keyGen.privateKey());
        success = privKeySaver.finalize();
        if (!success) {
            errorMessage = privKeySaver.errorString();
            return PrivKeyCreateError;
        }

        Utils::FileSaver pubKeySaver(pubKeyPath);
        publicKey = keyGen.publicKey();
        pubKeySaver.write(publicKey);
        success = pubKeySaver.finalize();
        if (!success) {
            errorMessage = pubKeySaver.errorString();
            return PubKeyCreateError;
        }

    } else {
        Utils::FileReader pubKeyReader;
        success = pubKeyReader.fetch(pubKeyPath);
        if (!success) {
            errorMessage = pubKeyReader.errorString();
            return PubKeyReadError;
        }

        publicKey = pubKeyReader.data();

        Utils::FileReader authKeysReader;
        success = authKeysReader.fetch(authorizedKeysPath);
        if (!success) {
            errorMessage = authKeysReader.errorString();
            return AuthKeysReadError;
        }

        if (authKeysReader.data().contains(publicKey)) {
            return NoError;
        }
    }

    Utils::FileSaver authorizedKeysSaver(authorizedKeysPath, QIODevice::Append);
    authorizedKeysSaver.write(publicKey);
    success = authorizedKeysSaver.finalize();
    if (!success) {
        errorMessage = authorizedKeysSaver.errorString();
        return AuthKeysAppendError;
    }

    return NoError;
}

QByteArray MerConnectionManager::readAllStandardOutput()
{
    return m_sshRemoteProcessRunner->readAllStandardOutput();
}

QByteArray MerConnectionManager::readAllStandardError()
{
    return m_sshRemoteProcessRunner->readAllStandardError();
}

} // Internal
} // Mer
