/****************************************************************************
**
** Copyright (C) 2012 Digia Plc and/or its subsidiary(-ies).
** Contact: http://www.qt-project.org/legal
**
** This file is part of Qt Creator.
**
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Digia.  For licensing terms and
** conditions see http://qt.digia.com/licensing.  For further information
** use the contact form at http://qt.digia.com/contact-us.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Digia gives you certain additional
** rights.  These rights are described in the Digia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
****************************************************************************/

#include "merdeployconfigurationfactory.h"

#include "merconstants.h"
#include "merdeployconfiguration.h"

#include <coreplugin/idocument.h>
#include <extensionsystem/pluginmanager.h>
#include <projectexplorer/buildstep.h>
#include <projectexplorer/buildsteplist.h>
#include <projectexplorer/kitinformation.h>
#include <projectexplorer/projectexplorerconstants.h>
#include <projectexplorer/project.h>
#include <projectexplorer/projectnodes.h>
#include <projectexplorer/target.h>
#include <qtsupport/baseqtversion.h>
#include <qtsupport/qtkitinformation.h>

using namespace ProjectExplorer;

namespace Mer {
namespace Internal {

MerDeployConfigurationFactory::MerDeployConfigurationFactory(QObject *parent)
    : DeployConfigurationFactory(parent)
{
}

QList<Core::Id> MerDeployConfigurationFactory::availableCreationIds(Target *parent) const
{
    if (ProjectExplorer::DeviceTypeKitInformation::deviceTypeId(parent->kit()) == Constants::MER_DEVICE_TYPE)
        return QList<Core::Id>() << MerDeployConfiguration::CreateDeployConfigurationId;
    return QList<Core::Id>();
}

QString MerDeployConfigurationFactory::displayNameForId(const Core::Id id) const
{
    return MerDeployConfiguration::theDisplayName();
}

bool MerDeployConfigurationFactory::canCreate(Target *parent, const Core::Id id) const
{
    return availableCreationIds(parent).contains(id);
}

// copy & pasted from src/plugins/projectexplorer/processstep.cpp
namespace {
const char PROCESS_STEP_ID[] = "ProjectExplorer.ProcessStep";
const char PROCESS_COMMAND_KEY[] = "ProjectExplorer.ProcessStep.Command";
const char PROCESS_WORKINGDIRECTORY_KEY[] = "ProjectExplorer.ProcessStep.WorkingDirectory";
const char PROCESS_ARGUMENTS_KEY[] = "ProjectExplorer.ProcessStep.Arguments";
}
// copy & paste end

BuildStep *processStep(BuildStepList *parent, const QString &command, const QString &arguments, const QString &workingDirectory)
{
    IBuildStepFactory *processStepFactory = 0;
    const Core::Id processStepID(PROCESS_STEP_ID);
    const QList<IBuildStepFactory*> factories = ExtensionSystem::PluginManager::getObjects<IBuildStepFactory>();
    foreach (IBuildStepFactory *factory, factories) {
        if (factory->canCreate(parent, processStepID)) {
            processStepFactory = factory;
            break;
        }
    }

    Q_ASSERT(processStepFactory); // Has ID of ProcessStepFactory changed?

    BuildStep *step = processStepFactory->create(parent, processStepID);
    QVariantMap map = step->toMap();
    map.insert(QLatin1String(PROCESS_COMMAND_KEY), command);
    map.insert(QLatin1String(PROCESS_ARGUMENTS_KEY), arguments);
    map.insert(QLatin1String(PROCESS_WORKINGDIRECTORY_KEY), workingDirectory);
    step->fromMap(map);

    return step;
}

DeployConfiguration *MerDeployConfigurationFactory::create(Target *parent, const Core::Id id)
{
    Q_ASSERT(canCreate(parent, id));

    // TODO: find the proper way to create a (Custum) Process Step
    MerDeployConfiguration *dc = new MerDeployConfiguration(parent, id);
    if (!dc)
        return dc;

    QtSupport::BaseQtVersion *qtVersion = QtSupport::QtKitInformation::qtVersion(parent->kit());
    const QString wrapperScriptsDir = qtVersion->qmakeCommand().parentDir().toString();
    const QString projectDirectory = parent->project()->projectDirectory();
    const QString projectName = QFileInfo(parent->project()->document()->fileName()).baseName();
    const QString merTarget = parent->kit()->displayName(); // TODO: find better way for this
    const QString rpmPackage = QString::fromLatin1("rpm/%1-%2.rpm").arg(projectName).arg(merTarget);

    const QString specifyCommand = wrapperScriptsDir + QLatin1Char('/') + QLatin1String(Constants::MER_WRAPPER_SPECIFY);
    const QString specifyParams = QString::fromLatin1("rpm/%1.yaml").arg(projectName);
    BuildStep *specifyStep = processStep(dc->stepList(), specifyCommand, specifyParams, projectDirectory);
    dc->stepList()->insertStep(dc->stepList()->count(), specifyStep);

    const QString rmCommand = wrapperScriptsDir + QLatin1Char('/') + QLatin1String(Constants::MER_WRAPPER_RM);
    const QString rmParams = QLatin1String("-rf /home/mersdk/rpmbuild/RPMS/");
    BuildStep *rmStep = processStep(dc->stepList(), rmCommand, rmParams, projectDirectory);
    dc->stepList()->insertStep(dc->stepList()->count(), rmStep);

    const QString mbCommand = wrapperScriptsDir + QLatin1Char('/') + QLatin1String(Constants::MER_WRAPPER_MB);
    const QString mbParams = QString::fromLatin1("build -t %1 rpm/%2.spec").arg(merTarget).arg(projectName);
    BuildStep *mbStep = processStep(dc->stepList(), mbCommand, mbParams, projectDirectory);
    dc->stepList()->insertStep(dc->stepList()->count(), mbStep);

    const QString mvCommand = wrapperScriptsDir + QLatin1Char('/') + QLatin1String(Constants::MER_WRAPPER_MV);
    const QString mvParams = QLatin1String("-T /home/mersdk/rpmbuild/RPMS/*/* ") + rpmPackage;
    BuildStep *mvStep = processStep(dc->stepList(), mvCommand, mvParams, projectDirectory);
    dc->stepList()->insertStep(dc->stepList()->count(), mvStep);

    return dc;
}

} // Internal
} // Mer
