/****************************************************************************
**
** Copyright (C) 2012 Digia Plc and/or its subsidiary(-ies).
** Contact: http://www.qt-project.org/legal
**
** This file is part of Qt Creator.
**
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Digia.  For licensing terms and
** conditions see http://qt.digia.com/licensing.  For further information
** use the contact form at http://qt.digia.com/contact-us.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Digia gives you certain additional
** rights.  These rights are described in the Digia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
****************************************************************************/

#include "merkitmanager.h"
#include "merqtversion.h"
#include "mertoolchain.h"
#include "merconstants.h"
#include "vmsettings.h"
#include "merconnectionmanager.h"

#include <projectexplorer/toolchainmanager.h>
#include <debugger/debuggerkitinformation.h>
#include <qtsupport/qtkitinformation.h>
#include <qtsupport/qtversionmanager.h>
#include <qtsupport/qtversionfactory.h>
#include <utils/qtcassert.h>
#include <utils/environment.h>
#include <coreplugin/icore.h>
#include <ssh/sshconnection.h>
#include <extensionsystem/pluginmanager.h>

#include <QTextStream>
#include <QDesktopServices>
#include <QFile>
#include <QDir>
#include <QTimer>

namespace Mer {
namespace Internal {

const char MER_WRAPPER[] = ":/mer/scripts/merWrapper";

static const char* wrapperScriptCopies[] = {
    "qmake",
    "make",
    "gcc",
    "gdb",
    Constants::MER_WRAPPER_SPECIFY,
    Constants::MER_WRAPPER_MB,
    Constants::MER_WRAPPER_RM,
    Constants::MER_WRAPPER_MV
};

class MerKitManagerPrivate
{
public:
    MerKitManagerPrivate() : state(MerKitManager::Idle), qtVersion(0), toolChain(0), kit(0)
    {
        timer.setSingleShot(true);
        timer.setInterval(0);
    }

    QTimer timer;
    QString targetName;
    QString sysRoot;
    MerKitManager::State state;

    QtSupport::BaseQtVersion *qtVersion;
    ProjectExplorer::GccToolChain *toolChain;
    ProjectExplorer::Kit *kit;

    VBoxManage vBoxManage;
};

inline QString targetToolsDirectory(const QString &targetName)
{
    return MerKitManager::devToolsDirectory() + targetName + QLatin1Char('/');
}

///////////////////////////////////////////////////////////////////////
//
// MerKitManager
//
///////////////////////////////////////////////////////////////////////

MerKitManager *MerKitManager::m_kitManager = 0;

MerKitManager::MerKitManager()
    : d(new MerKitManagerPrivate)
{
    connect(&d->timer, SIGNAL(timeout()), this, SLOT(changeState()));
    connect(&d->vBoxManage, SIGNAL(progress(QString,int,bool)),
            this, SLOT(vBoxProgress(QString,int,bool)));
}

MerKitManager::~MerKitManager()
{
    delete d;
}

MerKitManager *MerKitManager::instance()
{
    if (m_kitManager)
        return m_kitManager;

    m_kitManager = new MerKitManager;
    return m_kitManager;
}

void MerKitManager::createKit(const QString &targetName, const QString &sysRoot)
{
    if (d->state != Idle)
        return;
    d->targetName = targetName;
    d->sysRoot = sysRoot;
    d->state = CreateKit;
    changeState();
}

void MerKitManager::removeKit(const QString &targetName)
{
    if (d->state != Idle)
        return;
    d->kit = ProjectExplorer::KitManager::instance()->find(Core::Id(targetName));
    if (!d->kit) {
        emit progress(tr("Failed to find kit for %1!").arg(targetName), Done, false);
    } else {
        d->targetName = targetName;
        d->state = RemoveKit;
        changeState();
    }
}

void MerKitManager::updateVMConfig(const VMInfo &info)
{
    QMap<QString, QString> map;
    map.insert(QLatin1String("SDK_SSH_PORT"), QString::number(info.sshPort));
    map.insert(QLatin1String("MER_SHARED_HOME"), info.sharedHome);
    updateVMConfig(map);
}

QString MerKitManager::devToolsDirectory()
{
    QFileInfo settingsLocation(ExtensionSystem::PluginManager::settings()->fileName());
    return settingsLocation.absolutePath() + QLatin1String("/mer-dev-tools/");
}

void MerKitManager::updateVMConfig(const QMap<QString, QString> &map)
{
    QString targetPath = devToolsDirectory();
    QDir targetDir(targetPath);
    if (!targetDir.exists() && !targetDir.mkpath(targetPath))
        return;

    Utils::FileSaver vmConfigSaver(targetPath + QLatin1String("mer_vm.config"));
    QTextStream o(vmConfigSaver.file());
    o << "# Mer SDK VM config file\n";
    o << "# This file is overwritten by QtCreator ";
    o << "when the Mer SDK VM configuration is changed\n";
    o << "SDK_SSH_USER=" << Constants::MER_SDK_USERNAME << "\n";
    o << "SDK_SSH_KEYFILE=" <<
         MerConnectionManager::instance()->sshConnectionParameters().privateKeyFile << "\n";
    QMap<QString, QString>::const_iterator i = map.constBegin();
    while (i != map.constEnd()) {
        o << i.key() << "=" << i.value() << endl;
        ++i;
    }
    vmConfigSaver.finalize();
}

void MerKitManager::setWorkingDirectory(const QString &workingDirectory)
{
    QString targetPath = devToolsDirectory();
    QDir targetDir(targetPath);
    if (!targetDir.exists() && !targetDir.mkpath(targetPath))
        return;

    Utils::FileSaver buildConfigSaver(targetPath + QLatin1String("build_dir.config"));
    QTextStream o(buildConfigSaver.file());
    o << "# Mer SDK build config file\n";
    o << "# This file is overwritten by QtCreator\n";
    o << "BUILD_DIR=" << workingDirectory << "\n";
    buildConfigSaver.finalize();
}

void MerKitManager::createScripts()
{

    bool ok = true;

    QString targetPath = targetToolsDirectory(d->targetName);
    QDir targetDir(targetPath);
    if (!targetDir.exists() && !targetDir.mkpath(targetPath))
        ok = false;

    if (ok) {
        QFile::Permissions rwrw = QFile::ReadOwner|QFile::ReadUser|QFile::ReadGroup
                |QFile::WriteOwner|QFile::WriteUser|QFile::WriteGroup;
        QFile::Permissions rwxrwx = rwrw|QFile::ExeOwner|QFile::ExeUser|QFile::ExeGroup;
        for (size_t i = 0; i < sizeof(wrapperScriptCopies) / sizeof(wrapperScriptCopies[0]); ++i) {
            const QString scriptCopy = targetPath + QLatin1String(wrapperScriptCopies[i]);
            ok = QFile::copy(QLatin1String(MER_WRAPPER), scriptCopy);
            if (!ok)
                break;
            QFile::setPermissions(scriptCopy, rwxrwx);
        }
    }

    if (ok) {
        Utils::FileSaver targetConfigSaver(targetPath + QLatin1String("mer_target.config"));
        QTextStream o(targetConfigSaver.file());
        o << "# Mer SDK target config file\n";
        o << "# This file is overwritten by QtCreator";
        o << "when the Mer Target configuration is changed\n";
        o << "MER_SB2_TARGET='" << d->targetName <<"'\n";
        o << "MER_SYSROOT='" << d->sysRoot <<"'\n";
        ok = targetConfigSaver.finalize();
    }

    if (!ok)
        d->state = ScriptsCreateError;
    else
        d->state = ScriptsCreated;
    changeState();
}

bool MerKitManager::removeScripts()
{
    // delete the target directory
    return Utils::FileUtils::removeRecursively(targetToolsDirectory(d->targetName));
}

void MerKitManager::addQtVersion()
{
    Utils::FileName qmake = Utils::FileName::fromString(targetToolsDirectory(d->targetName) +
                                                        QLatin1String("qmake"));
    // Is there a qtversion present for this qmake?
    d->qtVersion = QtSupport::QtVersionManager::instance()->qtVersionForQMakeBinary(qmake);
    if (!d->qtVersion) {
        d->qtVersion = QtSupport::QtVersionFactory::createQtVersionFromQMakePath(qmake, true,
                                                                                 d->targetName);
        if (!d->qtVersion) {
            d->state = AddQtVersionError;
        }
        else {
            QtSupport::QtVersionManager::instance()->addVersion(d->qtVersion);
            d->state = QtVersionAdded;
        }
    } else {
        d->state = QtVersionAdded;
    }
    changeState();
}

void MerKitManager::removeQtVersion()
{
    QTC_ASSERT(d->kit, return);
    QtSupport::BaseQtVersion *qtVersion = QtSupport::QtKitInformation::qtVersion(d->kit);
    if (qtVersion)
        QtSupport::QtVersionManager::instance()->removeVersion(qtVersion);
}

void MerKitManager::registerToolChain()
{
    QString gcc = targetToolsDirectory(d->targetName) + QLatin1String("gcc");
    d->toolChain = new MerToolChain(true);
    d->toolChain->setDisplayName(d->toolChain->typeDisplayName() +
                                 QString(QLatin1String(" %1")).arg(d->targetName));
    d->toolChain->setCompilerCommand(Utils::FileName::fromString(gcc));
    if (!ProjectExplorer::ToolChainManager::instance()->registerToolChain(d->toolChain))
        d->state = RegisterToolChainError;
    else
        d->state = ToolChainRegistered;
    changeState();
}

void MerKitManager::deregisterToolChain()
{
    QTC_ASSERT(d->kit, return);
    ProjectExplorer::ToolChain *tc = ProjectExplorer::ToolChainKitInformation::toolChain(d->kit);
    if (tc)
        ProjectExplorer::ToolChainManager::instance()->deregisterToolChain(tc);
}

void MerKitManager::registerKit()
{
    QString gdb = targetToolsDirectory(d->targetName) + QLatin1String("gdb");
    ProjectExplorer::Kit *k = new ProjectExplorer::Kit(Core::Id(d->targetName));
    k->setDisplayName(tr("%1").arg(d->targetName));
    k->setIconPath(QLatin1String(Constants::MER_OPTIONS_CATEGORY_ICON));

    QtSupport::QtKitInformation::setQtVersion(k, d->qtVersion);
    ProjectExplorer::ToolChainKitInformation::setToolChain(k, d->toolChain);
    ProjectExplorer::SysRootKitInformation::setSysRoot(k, Utils::FileName::fromString(d->sysRoot));
    ProjectExplorer::DeviceTypeKitInformation::setDeviceTypeId(k, Constants::MER_DEVICE_TYPE);
    Debugger::DebuggerKitInformation::setDebuggerCommand(k, Utils::FileName::fromString(gdb));
    Debugger::DebuggerKitInformation::setEngineType(k, Debugger::GdbEngineType);
    if (!ProjectExplorer::KitManager::instance()->registerKit(k))
        d->state = RegisterKitError;
    else
        d->state = KitRegistered;
    changeState();
}

void MerKitManager::deregisterKit()
{
    QTC_ASSERT(d->kit, return);
    ProjectExplorer::KitManager::instance()->deregisterKit(d->kit);
}

void MerKitManager::changeState()
{
    switch (d->state) {
    case CreateKit:
        emit progress(tr("Checking VM..."), CheckVM, false);
        d->state = CheckVM;
        d->timer.start();
        break;
    case RemoveKit:
        emit progress(tr("Deleting Scripts..."), RemoveScripts, false);
        d->state = RemoveScripts;
        d->timer.start();
        break;
    case CheckVM:
        checkVM();
        break;
    case VMIsRunning:
        emit progress(tr("Creating Scripts..."), CreateScripts, false);
        d->state = CreateScripts;
        d->timer.start();
        break;
    case VMRunError:
        emit progress(tr("VM is not running!"), Done, false);
        d->state = Done;
        d->timer.start();
        break;
    case CreateScripts:
        createScripts();
        break;
    case ScriptsCreated:
        emit progress(tr("Adding Qt version..."), AddQtVersion, false);
        d->state = AddQtVersion;
        d->timer.start();
        break;
    case ScriptsCreateError:
        emit progress(tr("Failed to create scripts!"), Done, false);
        removeScripts();
        d->state = Done;
        d->timer.start();
        break;
    case AddQtVersion:
        addQtVersion();
        break;
    case QtVersionAdded:
        emit progress(tr("Registering tool chain..."), RegisterToolChain, false);
        d->state = RegisterToolChain;
        d->timer.start();
        break;
    case AddQtVersionError:
        emit progress(tr("Failed to add Qt Version!"), Done, false);
        removeScripts();
        d->state = Done;
        d->timer.start();
        break;
    case RegisterToolChain:
        registerToolChain();
        break;
    case ToolChainRegistered:
        emit progress(tr("Registering Kit..."), RegisterKit, false);
        d->state = RegisterKit;
        d->timer.start();
        break;
    case RegisterToolChainError:
        emit progress(tr("Failed to register tool chain!"), Done, false);
        QtSupport::QtVersionManager::instance()->removeVersion(d->qtVersion);
        removeScripts();
        d->state = Done;
        d->timer.start();
        break;
    case RegisterKit:
        registerKit();
        break;
    case KitRegistered:
        emit progress(tr("Successfully added target."), Done, true);
        d->state = Done;
        d->timer.start();
        break;
    case RegisterKitError:
        emit progress(tr("Failed to register kit!"), Done, false);
        ProjectExplorer::ToolChainManager::instance()->deregisterToolChain(d->toolChain);
        QtSupport::QtVersionManager::instance()->removeVersion(d->qtVersion);
        removeScripts();
        d->state = Done;
        d->timer.start();
        break;
    case RemoveScripts:
        removeScripts();
        emit progress(tr("Removing Qt version..."), RemoveQtVersion, false);
        d->state = RemoveQtVersion;
        d->timer.start();
        break;
    case RemoveQtVersion:
        removeQtVersion();
        emit progress(tr("Deregistering tool chain..."), DeregisterToolChain, false);
        d->state = DeregisterToolChain;
        d->timer.start();
        break;
    case DeregisterToolChain:
        deregisterToolChain();
        emit progress(tr("Deregistering kit..."), DeregisterKit, false);
        d->state = DeregisterKit;
        d->timer.start();
        break;
    case DeregisterKit:
        deregisterKit();
        emit progress(tr("Successfully removed target."), Done, true);
        d->state = Done;
        d->timer.start();
        break;
    case Done:
    default:
        resetState();
    }
}

void MerKitManager::vBoxProgress(const QString &msg, int progressValue, bool success)
{
    emit progress(msg, CheckVM, success);
    if (progressValue == VBoxManage::Done) {
        if (success)
            d->state = VMIsRunning;
        else
            d->state = VMRunError;
        changeState();
    }
}

void MerKitManager::resetState()
{
    d->state = Idle;
    d->qtVersion = 0;
    d->toolChain = 0;
    d->kit = 0;
}

void MerKitManager::checkVM()
{
    // Update VM Settings
    VMSettings vms;
    vms.fromSettings(Core::ICore::settings());
    d->vBoxManage.testSshConnection(vms.vmName);
}

} // namespace Internal
} // namespace Mer
