/****************************************************************************
**
** Copyright (C) 2012 Digia Plc and/or its subsidiary(-ies).
** Contact: http://www.qt-project.org/legal
**
** This file is part of Qt Creator.
**
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Digia.  For licensing terms and
** conditions see http://qt.digia.com/licensing.  For further information
** use the contact form at http://qt.digia.com/contact-us.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Digia gives you certain additional
** rights.  These rights are described in the Digia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
****************************************************************************/

#include "targetoptionspage.h"
#include "ui_targetoptionspage.h"
#include "merconstants.h"
#include "merkitmanager.h"

#include <coreplugin/icore.h>
#include <utils/qtcassert.h>

#include <QFileDialog>
#include <QFileInfo>
#include <QMessageBox>

namespace Mer {
namespace Internal {

///////////////////////////////////////////////////////////////////////
//
// TargetOptionsPage
//
///////////////////////////////////////////////////////////////////////

TargetOptionsPage::TargetOptionsPage(QObject *parent) :
    MerOptionsPage(parent)
{
    m_settings.fromSettings(Core::ICore::settings());
    setId(QLatin1String(Constants::TARGET_OPTIONS_ID));
    setDisplayName(QCoreApplication::translate("Mer", Constants::TARGET_OPTIONS_NAME));
}

QWidget *TargetOptionsPage::createPage(QWidget *parent)
{
    m_widget = new TargetOptionsWidget(parent);
    m_widget->setSettings(m_settings);
    if (m_searchKeyWords.isEmpty())
        m_searchKeyWords = m_widget->searchKeyWordMatchString();
    return m_widget;
}

void TargetOptionsPage::apply()
{
    if (m_widget) {
        const TargetSettings newSettings = m_widget->settings();
        if (newSettings != m_settings) {
            m_settings = newSettings;
            m_settings.toSettings(Core::ICore::settings());
        }
    }
}

bool TargetOptionsPage::matches(const QString &key) const
{
    return m_searchKeyWords.contains(key, Qt::CaseInsensitive);
}

///////////////////////////////////////////////////////////////////////
//
// TargetOptionsWidget
//
///////////////////////////////////////////////////////////////////////

TargetOptionsWidget::TargetOptionsWidget(QWidget *parent) :
    QWidget(parent),
    m_ui(new Ui::TargetOptionsPage),
    m_invalidTargetIcon(QLatin1String(":/mer/images/compile_error.png"))
{
    m_ui->setupUi(this);
    m_ui->merTargets->header()->setResizeMode(QHeaderView::ResizeToContents);
    m_ui->merTargets->header()->setStretchLastSection(false);
    m_ui->merTargets->setTextElideMode(Qt::ElideNone);
    m_ui->activityProgressBar->setMaximum(MerKitManager::Done);
    m_ui->activityProgressBar->setHidden(true);

    connect(m_ui->addButton, SIGNAL(clicked()), this, SLOT(addTarget()));
    connect(m_ui->removeButton, SIGNAL(clicked()), this, SLOT(removeTarget()));
    connect(MerKitManager::instance(), SIGNAL(progress(QString,int,bool)),
            this, SLOT(onProgress(QString,int,bool)));
    updateButtons();
}

TargetOptionsWidget::~TargetOptionsWidget()
{
    delete m_ui;
}

TargetSettings TargetOptionsWidget::settings() const
{
    TargetSettings s;
    QTreeWidgetItem *item = m_ui->merTargets->topLevelItem(0);
    while (item) {
        QVariantList data = QVariantList() << QVariant(item->icon(0).isNull())
                                           << QVariant(item->text(1));
        s.targets.insert(item->text(0), data);
        item = m_ui->merTargets->itemBelow(item);
    }
    return s;
}

void TargetOptionsWidget::setSettings(const TargetSettings &s)
{
    QHashIterator<QString, QVariantList> i(s.targets);
    while (i.hasNext()) {
        i.next();
        QTreeWidgetItem *item = new QTreeWidgetItem(m_ui->merTargets);
        item->setText(0, i.key());
        QVariantList data = i.value();
        item->setText(1, data.last().toString());
        item->setIcon(0, data.first().toBool() ? m_validTargetIcon : m_invalidTargetIcon);
        m_ui->merTargets->addTopLevelItem(item);
    }
    updateButtons();
}

QString TargetOptionsWidget::searchKeyWordMatchString() const
{
    const QChar blank = QLatin1Char(' ');
    QString keys;
    QTreeWidgetItem *item = m_ui->merTargets->topLevelItem(0);
    while (item) {
        keys += item->text(0) + blank;
        item = m_ui->merTargets->itemBelow(item);
    }

    return keys.trimmed();
}

void TargetOptionsWidget::onProgress(const QString &msg, int progress, bool success)
{
    m_ui->activityProgressBar->setValue(progress);
    m_ui->activityLabel->setText(msg);
    if (progress == MerKitManager::Done && success) {
        if (m_ui->merTargets->currentItem())
            m_ui->merTargets->currentItem()->setIcon(0, m_validTargetIcon);
    }
}

void TargetOptionsWidget::addTarget()
{
    // Default directory
    QLatin1String defaultDirectory("/srv/mer/targets");
#ifdef Q_OS_WIN
    defaultDirectory = QLatin1String();
#endif
    QString target = QFileDialog::getExistingDirectory(this, tr("Select target sysroot"),
                                                       defaultDirectory,
                                                       QFileDialog::ShowDirsOnly |
                                                       QFileDialog::DontResolveSymlinks);
    if (target.isEmpty())
        return;

    QDir targetDir = QDir(target);
    if (!targetDir.exists())
        return;

    if (!targetDir.entryList(QStringList() << QLatin1String("bin"), QDir::Dirs).count()) {
        QMessageBox::warning(this, tr("Directory not found"),
                             tr("The bin directory in %1 was not found.").arg(target));
        return;
    }

    QString targetName = targetDir.dirName();
    if (m_ui->merTargets->findItems(targetName, Qt::MatchExactly).count()) {
        QMessageBox::warning(this, tr("Target known"),
                             tr("This target is already registered as \"%1\".").arg(targetName));
        return;
    }

    m_ui->activityLabel->setText(QString());

    QTreeWidgetItem *item = new QTreeWidgetItem(m_ui->merTargets);
    item->setText(0, targetName);
    item->setIcon(0, m_invalidTargetIcon);
    item->setText(1, targetDir.absolutePath());
    m_ui->merTargets->addTopLevelItem(item);
    m_ui->merTargets->setCurrentItem(item);

    m_ui->activityProgressBar->setHidden(false);
    m_ui->activityProgressBar->setValue(0);
    MerKitManager::instance()->createKit(targetName, targetDir.absolutePath());
    updateButtons();
}

void TargetOptionsWidget::removeTarget()
{
    m_ui->activityProgressBar->setHidden(true);
    m_ui->activityLabel->setText(QString());

    QTreeWidgetItem *item = m_ui->merTargets->currentItem();
    if (item) {
        m_ui->activityProgressBar->setHidden(false);
        m_ui->activityProgressBar->setValue(0);
        MerKitManager::instance()->removeKit(item->text(0));
        delete item;
    }
    updateButtons();
}

void TargetOptionsWidget::updateButtons()
{
    if (m_ui->merTargets->topLevelItemCount())
        m_ui->removeButton->setEnabled(true);
    else
        m_ui->removeButton->setEnabled(false);
}

} // Internal
} // Mer
