#include "vboxmanage.h"
#include "merconnectionmanager.h"
#include "merkitmanager.h"

#include <utils/qtcassert.h>

#include <QProcess>
#include <QTimer>

const char VBOXMANAGE[] = "VBoxManage";
const char LIST[] = "list";
const char RUNNINGVMS[] = "runningvms";
const char VMS[] = "vms";
const char SHOWVMINFO[] = "showvminfo";
const char MACHINE_READABLE[] = "--machinereadable";
const char STARTVM[] = "startvm";

namespace Mer {
namespace Internal {

class VBoxManagePrivate
{
public:
    enum Command
    {
        None,
        VMs,
        RunningVMs,
        ShowVMInfo,
        StartVM
    };

    QProcess process;
    QString processOutput;
    QString processError;
    QString vmName;
    Command command;

    QSsh::SshConnection *connection;
    QSsh::SshConnectionParameters params;
    VBoxManage::State state;
    int tries;
    VMInfo vmInfo;
    bool stateMachine;
};

VBoxManage::VBoxManage(QObject *parent) :
    QObject(parent),
    d(new VBoxManagePrivate())
{
    d->command = VBoxManagePrivate::None;
    d->state = Idle;
    d->connection = 0;
    d->tries = 0;
    d->stateMachine = false;
    connect(&d->process, SIGNAL(finished(int)), this, SLOT(processFinished()));
    connect(&d->process, SIGNAL(error(QProcess::ProcessError)), this, SLOT(processError()));
    connect(&d->process, SIGNAL(readyReadStandardOutput()), this, SLOT(readyReadStandardOut()));
    connect(&d->process, SIGNAL(readyReadStandardError()), this, SLOT(readyReadStandardError()));
}

VBoxManage::~VBoxManage()
{
    closeConnection();
    delete d;
}

void VBoxManage::testSshConnection(const QString &vmName)
{
    if (d->state != Idle)
        return;
    emit progress(tr("Checking if VM is registered..."), CheckRegistered, false);
    d->vmName = vmName;
    d->state = CheckRegistered;
    d->stateMachine = true;
    isVMRegistered();
}

bool VBoxManage::isVMRunning(const QString &vmName)
{
    d->vmName = vmName;
    d->stateMachine = false;
    d->state = Idle;
    return isVMRunning();
}

bool VBoxManage::isVMRegistered(const QString &vmName)
{
    d->vmName = vmName;
    d->stateMachine = false;
    d->state = Idle;
    return isVMRegistered();
}

bool VBoxManage::getVMInfo(const QString &vmName)
{
    d->vmName = vmName;
    d->stateMachine = false;
    d->state = Idle;
    return getVMInfo();
}

bool VBoxManage::startVM(const QString &vmName)
{
    // Start VM if it is not running
    d->vmName = vmName;
    d->stateMachine = false;
    d->state = CheckRunning;
    return isVMRunning();
}

void VBoxManage::processFinished()
{
    switch (d->state) {
    case Idle: {
        switch (d->command) {
        case VBoxManagePrivate::VMs:
            emit vmRegistered(d->vmName, isVMListed());
            break;
        case VBoxManagePrivate::RunningVMs:
            emit vmRunning(d->vmName, isVMListed());
            break;
        case VBoxManagePrivate::ShowVMInfo:
            emit vmInfo(d->vmName, vmInfoFromOutput());
            break;
        default:
            break;
        }
        break;
    }
    case CheckRegistered:
        if (!isVMListed()) {
            emit progress(tr("VM is not registered!"), Done, false);
            d->state = Idle;
        } else {
            emit progress(tr("Fetching VM info..."), GetVMInfo, false);
            d->state = GetVMInfo;
            getVMInfo();
        }
        break;
    case GetVMInfo:
        d->vmInfo = vmInfoFromOutput();
        emit progress(tr("Checking if VM is running..."), CheckRunning, false);
        d->state = CheckRunning;
        isVMRunning();
        break;
    case CheckRunning:
        if (!isVMListed()) {
            emit progress(tr("VM is not running! Starting VM. (This will take a while)"),
                          Starting, false);
            startVM();
            if (d->stateMachine) {
                d->state = Starting;
                d->tries = 0;
                // Single shot timer to wait for lets say 12 seconds
                QTimer::singleShot(12000, this, SLOT(checkVMRunning()));
            } else {
                d->state = Idle;
            }
        } else {
            if (d->stateMachine) {
                emit progress(tr("Connecting to VM..."), Connecting, false);
                d->state = Connecting;
                connectToVM();
            }
        }
        break;
    case Starting:
        break;
    case Running:
        if (!isVMListed()) {
            // poll every 5s
            // try three times before giving up
            if (d->tries < 3) {
                d->tries++;
                emit progress(tr("VM is not running! Trying again(%1)...").arg(d->tries),
                              Running, false);
                QTimer::singleShot(5000, this, SLOT(checkVMRunning()));
            } else {
                emit progress(tr("VM is not running! Cannot connect."), Done, false);
                d->state = Idle;
            }
        } else {
            emit progress(tr("Connecting to VM..."), Connecting, false);
            d->state = Connecting;
            d->tries = 0;
            connectToVM();
        }
        break;
    case Connecting:
        break;
    default:
        break;
    }
}

void VBoxManage::checkVMRunning()
{
    emit progress(tr("Checking if VM is running..."), Running, false);
    d->state = Running;
    isVMRunning();
}

void VBoxManage::connectToVM()
{
    MerConnectionManager *connectionManager = MerConnectionManager::instance();
    QSsh::SshConnectionParameters params = connectionManager->sshConnectionParameters();
    params.port = d->vmInfo.sshPort;
    connectionManager->setSshConnectionParameters(params);
    QString errorMessage;
    if (connectionManager->createAndAuthorizeSshKeyIfNeeded(d->vmInfo.sharedHome, errorMessage) !=
            MerConnectionManager::NoError) {
        emit progress(errorMessage, Done, false);
        d->state = Idle;
        return;
    }
    d->connection = connectionManager->acquireConnection();

    connect(d->connection, SIGNAL(error(QSsh::SshError)), this, SLOT(onError(QSsh::SshError)));

    if (d->connection->state() == QSsh::SshConnection::Connected)
        onConnected();
    else
        connect(d->connection, SIGNAL(connected()), this, SLOT(onConnected()));
}

void VBoxManage::startVM()
{
    QStringList arguments;
    arguments.append(QLatin1String(STARTVM));
    arguments.append(d->vmName);
    QProcess::startDetached(QLatin1String(VBOXMANAGE), arguments);
}

bool VBoxManage::isVMRunning()
{
    if (d->process.state() != QProcess::NotRunning)
        return false;
    reset();
    QStringList arguments;
    arguments.append(QLatin1String(LIST));
    arguments.append(QLatin1String(RUNNINGVMS));
    d->process.start(QLatin1String(VBOXMANAGE), arguments);
    if (!d->process.waitForStarted())
        return false;
    d->command = VBoxManagePrivate::RunningVMs;
    return true;
}

bool VBoxManage::isVMRegistered()
{
    if (d->process.state() != QProcess::NotRunning)
        return false;
    reset();
    QStringList arguments;
    arguments.append(QLatin1String(LIST));
    arguments.append(QLatin1String(VMS));
    d->process.start(QLatin1String(VBOXMANAGE), arguments);
    if (!d->process.waitForStarted())
        return false;
    d->command = VBoxManagePrivate::VMs;
    return true;
}

bool VBoxManage::getVMInfo()
{
    if (d->process.state() != QProcess::NotRunning)
        return false;
    reset();
    QStringList arguments;
    arguments.append(QLatin1String(SHOWVMINFO));
    arguments.append(d->vmName);
    arguments.append(QLatin1String(MACHINE_READABLE));
    d->process.start(QLatin1String(VBOXMANAGE), arguments);
    if (!d->process.waitForStarted())
        return false;
    d->command = VBoxManagePrivate::ShowVMInfo;
    return true;
}

void VBoxManage::onConnected()
{
    emit progress(tr("Connection successful"), Done, true);
    closeConnection();
    d->state = Idle;
}

void VBoxManage::onError(QSsh::SshError error)
{
    // try three times before giving up
    if (error == QSsh::SshTimeoutError && d->tries < 3) {
        d->tries++;
        emit progress(tr("Could not connect to VM! Trying again(%1)...").arg(d->tries),
                      Connecting, false);
        QTimer::singleShot(5000, this, SLOT(connectToVM()));
    } else {
        emit progress(tr("Connection error: %1").arg(d->connection->errorString()),
                      Done, false);
        d->state = Idle;
    }
    closeConnection();
}

void VBoxManage::closeConnection()
{
    if (d->connection) {
        d->connection->disconnect(this);
        MerConnectionManager::instance()->releaseConnection(d->connection);
        d->connection = 0;
    }
}

void VBoxManage::processError()
{
}

void VBoxManage::readyReadStandardOut()
{
    d->processOutput.append(QString::fromLatin1(d->process.readAllStandardOutput()));
}

void VBoxManage::readyReadStandardError()
{
    d->processError.append(QString::fromLatin1(d->process.readAllStandardError()));
}

bool VBoxManage::isVMListed()
{
    return d->processOutput.indexOf(QRegExp(QString::fromLatin1("^\"%1\"").arg(d->vmName))) != -1;
}

void VBoxManage::reset()
{
    d->processError.clear();
    d->processOutput.clear();
}

VMInfo VBoxManage::vmInfoFromOutput()
{
    VMInfo vmInfo;
    vmInfo.sshPort = -1;

    // Get ssh port
    {
        int index = d->processOutput.indexOf(QLatin1String("Forwarding"));
        if (index == -1)
            return vmInfo;

        // Chop out the previous data;
        QString str = d->processOutput.mid(index);
        int startIndex = str.indexOf(QLatin1String("ssh"));
        if (startIndex == -1)
            return vmInfo;
        int endIndex = str.indexOf(QLatin1Char('\n'), startIndex + 1);
        QTC_ASSERT(endIndex != -1, return vmInfo);
        str = str.mid(startIndex, endIndex - startIndex + 1);
        QStringList args = str.split(QLatin1Char(','));
        // Name, Protocol, Host, Host Port, Guest IP, Guest port
        QTC_ASSERT(args.length() == 6, return vmInfo);
        vmInfo.sshPort = args.at(3).toUInt();
    }
    // Get shared home
    {
        int index = d->processOutput.indexOf(
                    QRegExp(QLatin1String("SharedFolderNameMachineMapping[0-9]+=")));
        if (index == -1)
            return vmInfo;

        // Chop out the previous data;
        QString str = d->processOutput.mid(index);
        int homeIndex = str.indexOf(QLatin1String("\"home\""));
        if (homeIndex == -1)
            return vmInfo;
        index = str.indexOf(
                    QRegExp(QLatin1String("SharedFolderPathMachineMapping[0-9]+=")), homeIndex + 1);
        if (index == -1)
            return vmInfo;
        int startIndex = str.indexOf(QLatin1Char('"'), index) + 1;
        int endIndex = str.indexOf(QLatin1Char('"'), startIndex + 1);
        QTC_ASSERT(startIndex != -1, return vmInfo);
        QTC_ASSERT(endIndex != -1, return vmInfo);
        vmInfo.sharedHome = str.mid(startIndex, endIndex - startIndex);
    }
    MerKitManager::updateVMConfig(vmInfo);
    return vmInfo;
}

} // Internal
} // Mer
